@@ -2,12 +2,14 @@
 
 namespace Icinga\Module\Businessprocess;
 
+use Exception;
+use Icinga\Application\Config;
 use Icinga\Exception\IcingaException;
 use Icinga\Exception\NotFoundError;
 use Icinga\Module\Businessprocess\Exception\NestingError;
 use Icinga\Module\Businessprocess\Modification\ProcessChanges;
+use Icinga\Module\Businessprocess\Storage\LegacyStorage;
 use Icinga\Module\Monitoring\Backend\MonitoringBackend;
-use Exception;
 
 class BpConfig
 {
@@ -29,6 +31,11 @@ class BpConfig
      */
     protected $backend;
 
+    /**
+     * @var LegacyStorage
+     */
+    protected $storage;
+
     /** @var  Metadata */
     protected $metadata;
 
@@ -51,7 +58,7 @@ class BpConfig
      *
      * @var int
      */
-    protected $state_type = self::HARD_STATE;
+    protected $state_type;
 
     /**
      * Warnings, usually filled at process build time
@@ -81,6 +88,20 @@ class BpConfig
      */
     protected $root_nodes = array();
 
+    /**
+     * Imported nodes
+     *
+     * @var ImportedNode[]
+     */
+    protected $importedNodes = [];
+
+    /**
+     * Imported configs
+     *
+     * @var BpConfig[]
+     */
+    protected $importedConfigs = [];
+
     /**
      * All host names { 'hostA' => true, ... }
      *
@@ -305,6 +326,26 @@ class BpConfig
         return $this;
     }
 
+    public function getStateType()
+    {
+        if ($this->state_type === null) {
+            if ($this->getMetadata()->has('Statetype')) {
+                switch ($this->getMetadata()->get('Statetype')) {
+                    case 'hard':
+                        $this->state_type = self::HARD_STATE;
+                        break;
+                    case 'soft':
+                        $this->state_type = self::SOFT_STATE;
+                        break;
+                }
+            } else {
+                $this->state_type = self::HARD_STATE;
+            }
+        }
+
+        return $this->state_type;
+    }
+
     public function useSoftStates()
     {
         $this->state_type = self::SOFT_STATE;
@@ -319,12 +360,12 @@ class BpConfig
 
     public function usesSoftStates()
     {
-        return $this->state_type === self::SOFT_STATE;
+        return $this->getStateType() === self::SOFT_STATE;
     }
 
     public function usesHardStates()
     {
-        return $this->state_type === self::HARD_STATE;
+        return $this->getStateType() === self::HARD_STATE;
     }
 
     public function addRootNode($name)
@@ -368,14 +409,32 @@ class BpConfig
      */
     public function getRootNodes()
     {
-        ksort($this->root_nodes);
+        if ($this->getMetadata()->isManuallyOrdered()) {
+            uasort($this->root_nodes, function (BpNode $a, BpNode $b) {
+                $a = $a->getDisplay();
+                $b = $b->getDisplay();
+                return $a > $b ? 1 : ($a < $b ? -1 : 0);
+            });
+        } else {
+            ksort($this->root_nodes, SORT_NATURAL | SORT_FLAG_CASE);
+        }
+
         return $this->root_nodes;
     }
 
     public function listRootNodes()
     {
         $names = array_keys($this->root_nodes);
-        sort($names);
+        if ($this->getMetadata()->isManuallyOrdered()) {
+            uasort($names, function ($a, $b) {
+                $a = $this->root_nodes[$a]->getDisplay();
+                $b = $this->root_nodes[$b]->getDisplay();
+                return $a > $b ? 1 : ($a < $b ? -1 : 0);
+            });
+        } else {
+            natcasesort($names);
+        }
+
         return $names;
     }
 
@@ -386,7 +445,14 @@ class BpConfig
 
     public function hasNode($name)
     {
-        return array_key_exists($name, $this->nodes);
+        if (array_key_exists($name, $this->nodes)) {
+            return true;
+        } elseif ($name[0] === '@') {
+            list($configName, $nodeName) = preg_split('~:\s*~', substr($name, 1), 2);
+            return $this->getImportedConfig($configName)->hasNode($nodeName);
+        }
+
+        return false;
     }
 
     public function hasRootNode($name)
@@ -397,12 +463,12 @@ class BpConfig
     public function createService($host, $service)
     {
         $node = new ServiceNode(
-            $this,
             (object) array(
                 'hostname' => $host,
                 'service'  => $service
             )
         );
+        $node->setBpConfig($this);
         $this->nodes[$host . ';' . $service] = $node;
         $this->hosts[$host] = true;
         return $node;
@@ -410,7 +476,8 @@ class BpConfig
 
     public function createHost($host)
     {
-        $node = new HostNode($this, (object) array('hostname' => $host));
+        $node = new HostNode((object) array('hostname' => $host));
+        $node->setBpConfig($this);
         $this->nodes[$host . ';Hoststatus'] = $node;
         $this->hosts[$host] = true;
         return $node;
@@ -434,9 +501,21 @@ class BpConfig
         return $this;
     }
 
-    public function listInvolvedHostNames()
+    public function listInvolvedHostNames(&$usedConfigs = null)
     {
-        return array_keys($this->hosts);
+        $hosts = $this->hosts;
+        if (! empty($this->importedNodes)) {
+            $usedConfigs[$this->getName()] = true;
+            foreach ($this->importedNodes as $node) {
+                if (isset($usedConfigs[$node->getConfigName()])) {
+                    continue;
+                }
+
+                $hosts += array_flip($node->getBpConfig()->listInvolvedHostNames($usedConfigs));
+            }
+        }
+
+        return array_keys($hosts);
     }
 
     /**
@@ -449,11 +528,12 @@ class BpConfig
      */
     public function createBp($name, $operator = '&')
     {
-        $node = new BpNode($this, (object) array(
+        $node = new BpNode((object) array(
             'name'        => $name,
             'operator'    => $operator,
             'child_names' => array(),
         ));
+        $node->setBpConfig($this);
 
         $this->addNode($name, $node);
         return $node;
@@ -482,14 +562,66 @@ class BpConfig
         }
 
         $node = new ImportedNode($this, $params);
+        $this->importedNodes[$node->getName()] = $node;
         $this->nodes[$node->getName()] = $node;
         return $node;
     }
 
+    public function getImportedNodes()
+    {
+        return $this->importedNodes;
+    }
+
+    public function getImportedConfig($name)
+    {
+        if (! isset($this->importedConfigs[$name])) {
+            $import = $this->storage()->loadProcess($name);
+
+            if ($this->usesSoftStates()) {
+                $import->useSoftStates();
+            } else {
+                $import->useHardStates();
+            }
+
+            $this->importedConfigs[$name] = $import;
+        }
+
+        return $this->importedConfigs[$name];
+    }
+
+    public function listInvolvedConfigs(&$configs = null)
+    {
+        if ($configs === null) {
+            $configs[$this->getName()] = $this;
+        }
+
+        foreach ($this->importedNodes as $node) {
+            if (! isset($configs[$node->getConfigName()])) {
+                $config = $node->getBpConfig();
+                $configs[$node->getConfigName()] = $config;
+                $config->listInvolvedConfigs($configs);
+            }
+        }
+
+        return $configs;
+    }
+
     /**
-     * @param $name
-     * @return Node
-     * @throws Exception
+     * @return LegacyStorage
+     */
+    protected function storage()
+    {
+        if ($this->storage === null) {
+            $this->storage = LegacyStorage::getInstance();
+        }
+
+        return $this->storage;
+    }
+
+    /**
+     * @param   string  $name
+     * @return  Node
+     * @throws  Exception
      */
     public function getNode($name)
     {
@@ -501,6 +633,11 @@ class BpConfig
             return $this->nodes[$name];
         }
 
+        if ($name[0] === '@') {
+            list($configName, $nodeName) = preg_split('~:\s*~', substr($name, 1), 2);
+            return $this->getImportedConfig($configName)->getNode($nodeName);
+        }
+
         // Fallback: if it is a service, create an empty one:
         $this->warn(sprintf('The node "%s" doesn\'t exist', $name));
         $pos = strpos($name, ';');
@@ -530,11 +667,12 @@ class BpConfig
         $this->calculateAllStates();
 
         $names = array_keys($this->getUnboundNodes());
-        $bp = new BpNode($this, (object) array(
+        $bp = new BpNode((object) array(
             'name' => '__unbound__',
             'operator' => '&',
             'child_names' => $names
         ));
+        $bp->setBpConfig($this);
         $bp->setAlias($this->translate('Unbound nodes'));
         return $bp;
     }
@@ -665,7 +803,16 @@ class BpConfig
             $nodes[$name] = $name === $alias ? $name : sprintf('%s (%s)', $alias, $node);
         }
 
-        natsort($nodes);
+        if ($this->getMetadata()->isManuallyOrdered()) {
+            uasort($nodes, function ($a, $b) {
+                $a = $this->nodes[$a]->getDisplay();
+                $b = $this->nodes[$b]->getDisplay();
+                return $a > $b ? 1 : ($a < $b ? -1 : 0);
+            });
+        } else {
+            natcasesort($nodes);
+        }
+
         return $nodes;
     }
 
@@ -768,7 +915,9 @@ class BpConfig
     {
         $args = func_get_args();
         array_shift($args);
-        $msg = vsprintf($msg, $args);
+        if (! empty($args)) {
+            $msg = vsprintf($msg, $args);
+        }
         if ($this->throwErrors) {
             throw new IcingaException($msg);
         }
@@ -837,4 +986,41 @@ class BpConfig
 
         return count($this->listBpNodes()) === 0;
     }
+
+    /**
+     * Export the config to array
+     *
+     * @param   bool    $flat   If false, children will be added to the array key children, else the array will be flat
+     *
+     * @return  array
+     */
+    public function toArray($flat = false)
+    {
+        $data = [
+            'name' => $this->getTitle(),
+            'path' => $this->getTitle()
+        ];
+
+        $children = [];
+
+        foreach ($this->getChildren() as $node) {
+            if ($flat) {
+                $children = array_merge($children, $node->toArray($data, $flat));
+            } else {
+                $children[] = $node->toArray($data, $flat);
+            }
+        }
+
+        if ($flat) {
+            $data = [$data];
+
+            if (! empty($children)) {
+                $data = array_merge($data, $children);
+            }
+        } else {
+            $data['children'] = $children;
+        }
+
+        return $data;
+    }
 }
