@@ -8,16 +8,20 @@ class ServiceNode extends MonitoredNode
 {
     protected $hostname;
 
+    /** @var string Alias of the host */
+    protected $hostAlias;
+
     protected $service;
 
     protected $className = 'service';
 
-    public function __construct(BpConfig $bp, $object)
+    protected $icon = 'service';
+
+    public function __construct($object)
     {
         $this->name = $object->hostname . ';' . $object->service;
         $this->hostname = $object->hostname;
         $this->service  = $object->service;
-        $this->bp       = $bp;
         if (isset($object->state)) {
             $this->setState($object->state);
         } else {
@@ -30,6 +34,30 @@ class ServiceNode extends MonitoredNode
         return $this->hostname;
     }
 
+    /**
+     * Get the host alias
+     *
+     * @return string
+     */
+    public function getHostAlias()
+    {
+        return $this->hostAlias;
+    }
+
+    /**
+     * Set the host alias
+     *
+     * @param string $hostAlias
+     *
+     * @return $this
+     */
+    public function setHostAlias($hostAlias)
+    {
+        $this->hostAlias = $hostAlias;
+
+        return $this;
+    }
+
     public function getServiceDescription()
     {
         return $this->service;
@@ -37,7 +65,7 @@ class ServiceNode extends MonitoredNode
 
     public function getAlias()
     {
-        return $this->hostname . ': ' . $this->service;
+        return $this->getHostAlias() . ': ' . $this->alias;
     }
 
     public function getUrl()
@@ -47,10 +75,10 @@ class ServiceNode extends MonitoredNode
             'service' => $this->getServiceDescription()
         );
 
-        if ($this->bp->hasBackendName()) {
-            $params['backend'] = $this->bp->getBackendName();
+        if ($this->getBpConfig()->hasBackendName()) {
+            $params['backend'] = $this->getBpConfig()->getBackendName();
         }
 
-        return Url::fromPath('monitoring/service/show', $params);
+        return Url::fromPath('businessprocess/service/show', $params);
     }
 }
